#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/serial.h>
#include <string.h>
#include <asm-generic/ioctls.h>
#include <unistd.h>
#include <sys/io.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define ACCESS_ONCE(x, y) \
{ \
	volatile char* tmp_xxx = &x; \
	*tmp_xxx = y; \
	usleep(10);\
}

#define ACCESS_GET(x) \
({ \
	volatile char* tmp_xxx = &x; \
	*tmp_xxx; \
})

#define F81504_CONFIG_SYSFS "/sys/class/tty/%s/device/config"
#define F81504_CONFIG_SYSFS_IOPORT "/sys/class/tty/%s/port"
#define F81504_PCI_SYSFS "/sys/class/tty/%s/device"

#define ARRAY_SIZE(arr)     (sizeof(arr) / sizeof(arr[0]))

#define BIT(x) (1ULL << (x))

static volatile unsigned char *can_mmio;

static int m_mem_fd = -1;
static unsigned char m_pci_path[64];

const static char *m_cStringArray_f81439[] = {
	"F81439(M0/M1/M2)",
	"\t0: 000 - RS422",
	"\t1: 001 - RS232",
	"\t2: 010 - RS485 Low Active",
	"\t3: 011 - RS485 High Active",
	"\t4: 100 - RS422 with termination",
	"\t5: 101 - Reserved",
	"\t6: 110 - RS485 Low Active with termination",
	"\t7: 111 - Shutdown",
	NULL,
};

const static int m_accpet_mode_f81439[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };

const static char *m_cStringArray_f81435[] = {
	"F81435(M0/M1/M2)",
	"\t0: 000 - RS422 High Active",
	"\t1: 001 - RS232",
	"\t2: 010 - RS485 High Active",
	"\t3: 011 - ccTALK",
	"\t4: 100 - RS422 Low Active",
	"\t5: 101 - RS485 Low Active (Auto Sensing)",
	"\t6: 110 - RS485 Low Active",
	"\t7: 111 - RS422 Low Active (Auto Sensing)",
	NULL,
};

const static int m_accpet_mode_f81435[] = { 0, 1, 2, 3, 4, 5, 6, 7, -1 };

const static char *m_cStringArray_f81433[] = {
	"F81433(M0/M1/R_DIS)",
#if 0
	"\t1: 001 - RS485 Low Active with termination",
	"\t2: 010 - RS422 with termination",
	"\t4: 100 - RS232",
	"\t5: 101 - RS485 Low Active",
	"\t6: 110 - RS422",
	"\t7: 111 - Shutdown",
#else
	"\t1: 001 - RS232",
	"\t2: 010 - RS422 with termination",
	"\t3: 011 - RS422",
	"\t4: 100 - RS485 Low Active with termination (Auto Sensing)",
	"\t5: 101 - RS485 Low Active (Auto Sensing)",
	"\t7: 111 - Shutdown",
#endif
	NULL,
};

const static int m_accpet_mode_f81433[] = { 1, 2, 3, 4, 5, 7, -1 };

typedef enum {
	transceiver_type_f81439,
	transceiver_type_f81435,
	transceiver_type_f81433,
	transceiver_type_invalid,
} transceiver_type;

const static char **m_cStringArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = m_cStringArray_f81439,
	[transceiver_type_f81435] = m_cStringArray_f81435,
	[transceiver_type_f81433] = m_cStringArray_f81433,
};

const static int *m_nAcceptArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = m_accpet_mode_f81439,
	[transceiver_type_f81435] = m_accpet_mode_f81435,
	[transceiver_type_f81433] = m_accpet_mode_f81433,
};

const static char *m_nNameArray[transceiver_type_invalid] = {
	[transceiver_type_f81439] = "F81439",
	[transceiver_type_f81435] = "F81435",
	[transceiver_type_f81433] = "F81433",
};

static int read_pci_byte(int addr, unsigned char *data)
{
	FILE *file;
	int len;

	file = fopen(m_pci_path, "r");
	if (!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fread(data, 1, 1, file);
	if (len != 1) {
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;
}

static int write_pci_byte(int addr, unsigned char data)
{
	FILE *file;
	int len;

	file = fopen(m_pci_path, "wb");
	if (!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fwrite(&data, 1, 1, file);
	if (len != 1) {
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;
}

static int write_pci_mask_byte(int addr, unsigned char mask, unsigned char data)
{
	int status;
	unsigned char tmp;

	status = read_pci_byte(addr, &tmp);
	if (status)
		return -1;

	tmp &= ~mask;
	tmp |= (mask & data);

	status = write_pci_byte(addr, tmp);
	if (status)
		return -1;

	return 0;
}

static void usage(char *argv[])
{
	const char **st;
	int i, j;

	fprintf(stderr, "Fintek F81504 set gpio tools\n");
	fprintf(stderr, "Set mode\n");
	fprintf(stderr, "\t%s <port> <mode>\n", argv[0]);
	fprintf(stderr, "Query mode\n");
	fprintf(stderr, "\t%s <port>\n", argv[0]);
	fprintf(stderr, "\nUart Mode table:\n");

	for (i = 0; i < transceiver_type_invalid; ++i) {
		st = m_cStringArray[i];

		for (j = 0; st[j]; ++j)
			fprintf(stderr, "%s\n", st[j]);

		fprintf(stderr, "\n");
	}

	fprintf(stderr, "\n");
}

static int is_f81504(void)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	for (len = 0; len < 4; ++len) {
		read_pci_byte(len, &buf[len]);
		//printf("%x", buf[len]);
	}

	//printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x11))
		return -1;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return 0;
	}

	return -1;
}

static int get_tty_idx(int io_addr)
{
	unsigned char buf[2];
	unsigned int i, j;
	int read_io;
	int len;

	for (i = 0; i < 12; ++i) {
		memset(buf, 0, sizeof(buf));

		for (j = 0; j < 2; ++j) {
			read_pci_byte(0x40 + 0x8 * i + 4 + 0, &buf[0]);
			read_pci_byte(0x40 + 0x8 * i + 4 + 1, &buf[1]);
		}

		read_io = (buf[1] << 8) | buf[0];

		if (io_addr == read_io)
			break;
	}

	if (i >= 12)
		return -1;

	return i;
}

static int read_ioport(char *tty)
{
	char buf[64];
	int len;
	int fd;
	int io_addr;

	sprintf(buf, F81504_CONFIG_SYSFS_IOPORT, tty);
	//printf("Target:%s\n", buf);

	fd = open(buf, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "open error %d\n", __LINE__);
		perror("");
		return -1;
	}

	memset(buf, 0, sizeof(buf));
	len = read(fd, buf, sizeof(buf));
	if (!len)
		return -1;

	io_addr = strtol(buf, NULL, 16);
	close(fd);

	return io_addr;
}

static int get_transceiver_type(char *tty, int idx, transceiver_type * type)
{
	unsigned char buf[128];
	FILE *file;
	int i, data[4];

	if (idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/transceiver", tty);

	file = fopen(buf, "r");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	for (i = 0; i < 4; ++i) {
		fscanf(file, "%d\n", &data[i]);
		//printf("%s: %d %d\n", __func__, i, data[i]);

		if (i == idx)
			*type = data[i];
	}

	fclose(file);
	return 0;
}

static int get_current_mode(char *tty, int idx, unsigned int *mode)
{
	unsigned char buf[128];
	FILE *file;
	int i, data[4];

	if (idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/gpio_ctrl", tty);

	file = fopen(buf, "r");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	for (i = 0; i < 4; ++i) {
		fscanf(file, "%d\n", &data[i]);
		//printf("%s: %d %d\n", __func__, i, data[i]);

		if (i == idx)
			*mode = data[i];
	}

	return 0;
}

static int check_transceiver_mode(transceiver_type type, int mode)
{
	const int *tb = m_nAcceptArray[type];
	int i;

	for (i = 0; tb[i] != -1; ++i) {
		if (mode == tb[i])
			return 0;
	}

	return -1;
}

static int change_gpio(char *tty, int idx, unsigned int mode)
{
	unsigned char buf[128];
	FILE *file;
	int len;

	if (idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/gpio_ctrl", tty);

	file = fopen(buf, "w");
	if (!file) {
		printf("open failed: %s\n", buf);
		return -1;
	}

	sprintf(buf, "%d %d\n", idx, mode);

	len = fprintf(file, buf, strlen(buf));
	if (len <= 0) {
		printf("set gpio failed: %d\n", len);
		return -1;
	}

	fclose(file);

	return 0;
}

int main(int argc, char *argv[])
{
	int length = 0;
	int i = 0, gpio_mode, devid, idx, is_fix;
	int ret = 0;
	int status;
	unsigned int io;
	struct serial_rs485 rs485conf;
	transceiver_type type;

	if (argc < 2) {
		usage(argv);
		return -1;
	}

	sprintf(m_pci_path, F81504_CONFIG_SYSFS, argv[1]);
	//printf("Target:%s\n", m_pci_path);

	status = is_f81504();
	if (status) {
		fprintf(stderr, "Not F81504 series\n");
		return status;
	}

	io = read_ioport(argv[1]);
	idx = get_tty_idx(io);

	//printf("%s: %s IO: %x\n", __func__, argv[1], io);
	//printf("%s: %s idx: %d\n", __func__, argv[1], idx);

	status = get_transceiver_type(argv[1], idx, &type);
	if (status) {
		printf("Cant get transeciver type\n");
		return status;
	}

	if (type == transceiver_type_f81433) {
		printf("Transceiver: F81433\n");
	} else if (type == transceiver_type_f81439) {
		printf("Transceiver: F81439\n");
	} else if (type == transceiver_type_f81435) {
		printf("Transceiver: F81435\n");
	} else {
		printf("Unknown\n");
	}

	status = get_current_mode(argv[1], idx, &gpio_mode);
	if (status) {
		printf("Cant get gpio mode\n");
		return status;
	}

	if (gpio_mode == 0xff) {
		// fix mode
		printf("current mode is fix mode\n");
		return 1;
	}

	printf("current mode: %d\n", gpio_mode);

	if (argc <= 2)
		return 0;

	gpio_mode = strtol(argv[2], NULL, 0);

	status = check_transceiver_mode(type, gpio_mode);
	if (status) {
		printf("%s not support mode: %d\n", m_nNameArray[type], gpio_mode);
		return 1;
	}

	status = change_gpio(argv[1], idx, gpio_mode);
	if (status) {
		printf("change gpio failed\n");
		return 1;
	}

	printf("set mode to: %d\n", gpio_mode);

	return 0;
}
