#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/serial.h>
#include <string.h>
#include <asm-generic/ioctls.h>

#define F81504_CONFIG_SYSFS "/sys/class/tty/%s/device/config"
#define F81504_CONFIG_SYSFS_IOPORT "/sys/class/tty/%s/port"


enum error {
	error_no_error = 0,

	error_parameter,
	error_open_failed,
	error_not_f81504,
	error_change_mode,
	error_read_id,
	error_read_io,
	error_read_pci,
	error_write_pci,
	error_io_not_found,
};

static char *m_cStringArray[] = {
	(char *)-1,
	"eModeRS232",
	"eModeRS485 - DE#/RE (TX Enable with RTS Low)",
	"eModeRS485_1 - DE/RE# (TX Enable with RTS High)",
	(char *)-1,
	NULL,
};

void usage(char *argv[])
{
	int i = 0;
	fprintf(stderr, "%s\n", argv[0]);
	fprintf(stderr, "%s <port> <mode>\n\n", argv[0]);
	fprintf(stderr, "\nUart Mode table:\n");

	for (i = 0; m_cStringArray[i]; ++i) {
		if (m_cStringArray[i] == (char *)-1)
			continue;
		fprintf(stderr, "%i --> %s\n", i, m_cStringArray[i]);
	}

	fprintf(stderr, "\n");
}

int is_f81504(int fd)
{
	int len;
	int tmp;
	char buf[5];

	memset(buf, 0, sizeof(buf));

	lseek(fd, 0, SEEK_SET);

	len = read(fd, buf, sizeof(buf));
	if (!len)
		return error_read_id;

	for (len = 0; len < 4; ++len)
		printf("%x", buf[len]);
	printf("\n");

	if ((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x11))
		return error_not_f81504;

	switch (buf[2]) {
	case 0x04:
	case 0x08:
	case 0x12:
		return error_no_error;
	}
	
	return error_not_f81504;
}

int change_mode(int fd, int mode, int io_addr)
{
	unsigned char buf[4];
	int read_io;
	int len;
	int i;

	for (i = 0; i < 12; ++i) {
		memset(buf, 0, sizeof(buf));
		lseek(fd, 0x40 + 0x8 * i + 4, SEEK_SET);

		len = read(fd, buf, 2);
		if (len != 2)
			return error_read_pci;

		read_io = (buf[1] << 8) | buf[0];

		if (io_addr == read_io )
			break;
	}

	if (i >= 12)
		return error_io_not_found;

	memset(buf, 0, sizeof(buf));
	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);

	len = read(fd, buf, 1);
	if (len != 1)
		return error_read_pci;

	switch (mode) {
	case 1:
		buf[0] = 0x01;
		break;
	case 2:
		buf[0] = 0x31;
		break;
	case 3:
		buf[0] = 0x11;
		break;
	default:
		return error_parameter;
	}

	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);
	len = write(fd, buf, 1);
	if (len != 1)
		return error_write_pci;
	
	return error_no_error;
}

int read_ioport(char* tty)
{
	char buf[64];
	int len;
	int fd;
	int io_addr;

	sprintf(buf, F81504_CONFIG_SYSFS_IOPORT, tty);
	printf("Target:%s\n", buf);

	fd = open(buf, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "open error %d\n", __LINE__);
		perror("");
		return error_open_failed;
	}	

	memset(buf, 0, sizeof(buf));
	len = read(fd, buf, sizeof(buf));
	if (!len)
		return error_read_io;

	io_addr = strtol(buf, NULL, 16);
	close(fd);

	return io_addr;
}

int main(int argc, char *argv[])
{
	int fd = -1;
	int length = 0;
	int i = 0;
	int ret = 0;
	char buf[128];
	int status;
	struct serial_rs485 rs485conf;

	if (argc < 3) {
		usage(argv);
		return error_parameter;
	}

	i = atoi(argv[2]);

	if (m_cStringArray[i] == (char *)-1) {
		usage(argv);
		return error_parameter;
	}


	sprintf(buf, F81504_CONFIG_SYSFS, argv[1]);
	printf("Target:%s\n", buf);

	fd = open(buf, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "open error %d\n", __LINE__);
		return error_open_failed;
	}

	status = is_f81504(fd);
	if (status) {
		fprintf(stderr, "Not F81504 series\n");
		close(fd);
		return status;
	}

	status = change_mode(fd, i, read_ioport(argv[1]));
	if (status) {
		fprintf(stderr, "change_mode failed\n");
		close(fd);
		return status;
	}

	printf("change to mode: %s\n", m_cStringArray[i]);
	close(fd);

	return error_no_error;
}
